"""
----------------------------------------------------------
vpm: utilities for parsing VirgoProcessMonitoring xml file
----------------------------------------------------------
"""

# B. Swinkels, 2014

from __future__ import print_function
from builtins import object
from xml.etree import ElementTree
from os import path, environ
from virgotools.pvt_utils import AttrDict


__all__ = ['VPM', 'vpm_status']

try:
    import virgotools.pvt_internal_defaults as defaults
except ImportError:
    import virgotools.pvt_external_defaults as defaults


VPM_ALIAS_KEY = "VPM_ALIAS"
DMS_FILE_KEY = "DMS_FILE"

class VPM(object):
    """
    Class for parsing the xml-file defining the processes in VPM.

    Parameters
    ----------
    xml_file : str
        filename of the xml file. Aliasses 'DAQ' and 'LLA' are defined.

    Attributes
    ----------
    apps : AttrDict
        mapping of application name to objects containing the application name,
        path to its executable and a list of processes instances
    procs : AttrDict
        mapping of process name to objects containing the process name, path to
        configuration file, the host on which it is running, its application and
        a flag to indicate if it is an optional process in VPM
    hosts : AttrDict
        mapping of host name to a list of processes running on that host

    Examples
    --------
    >>> vpm = VPM('DAQ')
    >>> vpm.procs.PrISYS
    >>> for proc in vpm.apps.Acl.procs:
    ...     print(proc.name)
    SIB2_Photodiodes
    SIB2_Quadrants
    [...]
    >>> vpm.procs.LSC_Pr.config
    '/virgoData/VirgoOnline/LSC_Pr.cfg'
    >>> for proc in vpm.hosts.rtpc3:
    ...     print(proc.name)
    ISC_Tpro
    LSC_Pr
    CaNoise
    ISC_Fb
    """
    def __init__(self, xml_file):

        #Check the enironmental variable
        VPM_alias_env = environ.get(VPM_ALIAS_KEY)

        #Parse the environmental variable
        if VPM_alias_env:
            if ":" not in VPM_alias_env:
                raise NameError("The string in %s environmental varible is not correctly formatted, the correct format is 'alias: xml_file'")
            VPM_alias = {}
            VPM_alias_env = VPM_alias_env.split(",")
            for ifile in VPM_alias_env:
                key,value = ifile.split(":")
                VPM_alias[key] = value
        elif defaults.VPM_alias:
            VPM_alias = defaults.VPM_alias
        else:
            VPM_alias = {}

        xml_file = VPM_alias.get(xml_file,xml_file)
        root = ElementTree.parse(xml_file).getroot()
        self.name = root.attrib['name']
        # self._root = root

        parent = {c: p for p in root.iter() for c in p}

        def lookup(elem, key, default=None):
            """looks up attrib on elem and all its parents"""
            while True:
                try:
                    return elem.attrib[key]
                except KeyError:
                    try:
                        elem = parent[elem]  # go to parent
                    except KeyError:  # reached root of tree
                        return default

        app_dir = root.get('application-dir',
                           '{{INSTALL_ROOT}}/{{PACKAGE}}/{{VERSION}}/{{TAG}}')
        alias = {  # global aliases
            '{{INSTALL_ROOT}}': root.get('install-root', '/virgoApp'),
            '{{DATA_ROOT}}': root.get('data-root', '/virgoData'),
            '{{LOG_ROOT}}': root.get('log-root', '/virgoLog'),
            '{{VPM}}': self.name.replace(' ', ''),
        }

        host_alias = {elem.attrib['name']: elem.attrib['target']
                      for elem in root.iter('host')}

        self.apps = AttrDict()
        for elem in root.iter('application'):
            app = AttrDict()
            name = elem.attrib['name']
            app.name = name
            try:
                p = elem.attrib['path']
            except KeyError:
                p = path.join(app_dir, name + elem.get('suffix', '.exe'))

            p = p.replace('{{INSTALL_ROOT}}',
                          lookup(elem, 'install-root', '/virgoApp'))
            p = p.replace('{{PACKAGE}}', elem.get('package', name))
            p = p.replace('{{VERSION}}', elem.get('version', '{{UNKNOWN_VERSION}}'))
            p = p.replace('{{TAG}}', elem.get('tag', '$UNAME'))
            app.path = path.normpath(p)
            app.procs = []
            self.apps[name] = app

        self.procs = AttrDict()
        self.hosts = AttrDict()
        for elem in root.iter('process'):
            proc = AttrDict()
            name = elem.get('name')
            proc.name = name

            app_name = elem.get('application')
            try:
                app = self.apps[app_name]

            except KeyError:
                print('Process {} has unknown application {}, skipping ...'.format(
                    name, app_name))
                continue  # skip this process, since it crashes elsewhere
            proc.app = app
            app.procs.append(proc)

            host = lookup(elem, 'host')
            host = host_alias.get(host, host)  # substitute aliases
            proc.host = host
            try:
                proc.config = elem.attrib['configuration']
            except KeyError:
                conf_dir = lookup(elem, 'configuration-dir',
                                  '{{DATA_ROOT}}/{{VPM}}')
                for k, v in alias.items():
                    conf_dir = conf_dir.replace(k, v)
                conf_dir = conf_dir.replace('{{APPLICATION}}', proc.app.name)
                proc.config = path.join(conf_dir, name + '.cfg')
            proc.optional = elem.attrib.get('optional') == 'true'

            self.procs[name] = proc
            self.hosts.setdefault(host, []).append(proc)





def vpm_status(DMS_file=defaults.DMS_file):
    """
    Get dict with status of all non-optional processes in VPM, by parsing
    output for DMS.

    Returns
    -------
    AttrDict
        mapping from Cm-name to process status with following code:
            0: everything ok
            1: error
            2: warning
    """
    DMS_file = environ.get(DMS_FILE_KEY, DMS_file)
    if not DMS_file:
        raise NameError("The default DMS file is not defined, please set the %s environmental variable or set the 'DMS_file' variable in %s/pvt_external_defaults.py" %(DMS_FILE_KEY,path.dirname(path.abspath(__file__))))

    with open(DMS_file) as fid:
        lines = fid.readlines()

    lines = ''.join(lines[2:-2])  # skip PHP stuff at begin and end
    DMS = ElementTree.fromstring(lines)

    proc_status = AttrDict()
    for elem in DMS.iter('qcalgo_subflag'):
        cm_name = elem.find('flag_chname').text
        status = elem.find('flag_status').text
        proc_status[cm_name] = int(status)

    # add gps time. Note comment in xml that it might be off by few second!
    proc_status.gps = int(DMS.find('qcheader').find('qcframe_gps').text)
    return proc_status
