import pytest
from mock import Mock
from unittest.mock import patch
from os import path, environ

from virgotools.cm_lib import CmError
from virgotools import interface_DSP

satserver_config_file = path.join(path.dirname(__file__), "TEST_SAT_GAINS.ini")

CM_NAME = "test_DSP"
TIMEOUT = interface_DSP.TIMEOUT
RAMPTIME = 4.0

mock_cm_send_receive = Mock()
interface_DSP.cm_send_receive = mock_cm_send_receive


@pytest.fixture
@patch.dict(environ,{"SATSERVER_CONFIG_FILE":satserver_config_file})
def dsp():
    # reset mock each time since it is used globally
    mock_cm_send_receive.reset_mock(return_value=True, side_effect=True)
    return interface_DSP.DSP("TWR", RAMPTIME)

@pytest.mark.skipif(interface_DSP.defaults.server=="",reason="The default server is not defined the defaults file")
def test_unknown_dsp():
    with pytest.raises(ValueError):
        dsp = interface_DSP.DSP("NONEXISTING")

@pytest.mark.skipif(interface_DSP.defaults.server=="",reason="The default server is not defined the defaults file")
def test_top_bottom(dsp):
    assert dsp._gname2crate["TOP"] == "Sa_TWR"
    assert dsp._gname2crate["BOTTOM"] == "Sc_TWR"

@pytest.mark.skipif(interface_DSP.defaults.server=="",reason="The default server is not defined the defaults file")
def test_nominal_get(dsp):
    # from top stage
    mock_cm_send_receive.return_value = [1.23]
    assert dsp.TOP == 1.23
    mock_cm_send_receive.assert_called_with(
        "get_value_reply", TIMEOUT, "SatServer", "get_value", "Sa_TWR", "TOP"
    )
    # from bottom stage
    mock_cm_send_receive.return_value = [4.56]
    assert dsp.BOTTOM == 4.56
    mock_cm_send_receive.assert_called_with(
        "get_value_reply", TIMEOUT, "SatServer", "get_value", "Sc_TWR", "BOTTOM"
    )


@pytest.mark.skipif(interface_DSP.defaults.server=="",reason="The default server is not defined the defaults file")
def test_bad_get(dsp):
    mock_cm_send_receive.return_value = [float("NaN")]
    with pytest.raises(interface_DSP.DSPGetError):
        dsp.BOTTOM

    mock_cm_send_receive.side_effect = CmError("Timeout")
    with pytest.raises(interface_DSP.DSPGetError):
        dsp.BOTTOM

@pytest.mark.skipif(interface_DSP.defaults.server=="",reason="The default server is not defined the defaults file")
def test_nominal_set(dsp):
    mock_cm_send_receive.return_value = ["OK"]
    dsp.BOTTOM = 3
    mock_cm_send_receive.assert_called_with(
        "set_value_reply",
        TIMEOUT,
        "SatServer",
        "set_value",
        "Sc_TWR",
        "BOTTOM",
        3.0,
        RAMPTIME,
    )

@pytest.mark.skipif(interface_DSP.defaults.server=="",reason="The default server is not defined the defaults file")
def test_bad_set(dsp):
    mock_cm_send_receive.return_value = ["some error string"]
    with pytest.raises(interface_DSP.DSPSetError):
        dsp.BOTTOM = 3

    mock_cm_send_receive.side_effect = CmError("Timeout")
    with pytest.raises(interface_DSP.DSPSetError):
        dsp.BOTTOM = 3

@pytest.mark.skipif(interface_DSP.defaults.server=="",reason="The default server is not defined the defaults file")
def test_unknown_gname(dsp):
    with pytest.raises(AttributeError):
        dsp.NONEXISTING
    with pytest.raises(AttributeError):
        dsp.NONEXISTING = 1.23
    with pytest.raises(ValueError):
        dsp.get_attribute("NONEXISTING")
    with pytest.raises(ValueError):
        dsp.ramp_attribute("NONEXISTING", 1.23, 2)
    mock_cm_send_receive.assert_not_called()

@pytest.mark.skipif(interface_DSP.defaults.server=="",reason="The default server is not defined the defaults file")
@pytest.mark.parametrize("value", ["1.23", float("nan"), float("inf"), -float("inf")])
def test_bad_input(dsp, value):
    mock_cm_send_receive.return_value = ["OK"]
    with pytest.raises(Exception):
        dsp.BOTTOM = value
    mock_cm_send_receive.assert_not_called()
